package com.ibm.xml.sdo.test;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.List;

import junit.framework.Assert;

import org.apache.jsr235.tck.util.XMLCompareUtil;
import org.junit.Before;
import org.junit.Test;

import com.ibm.xml.sdo.api.SDOUtil;
import commonj.sdo.DataObject;
import commonj.sdo.Property;
import commonj.sdo.Type;
import commonj.sdo.helper.HelperContext;

/**
 * Duplicate names tests
 */
public class DuplicateNamesTest {
    /** IBM binary license notice. */
    final static String IBM_COPYRIGHT = "Licensed Materials - Property of IBM\n"
            + "\n"
            + "YOR7-2004-0083\n"
            + "\n"
            + "(c) Copyright IBM Corp. 2004, 2005. All Rights Reserved.\n"
            + "\n"
            + "US Government Users Restricted Rights - Use, duplication or disclosure\n"
            + "restricted by GSA ADP Schedule Contract with IBM Corp.";

    private String RES_PATH_DUPLICATESXSD = "/DuplicateNamesTest/DuplicateNames.xsd";
    private String RES_PATH_TWOFOOSXML = "/DuplicateNamesTest/TwoFoos.xml";
    private String RES_PATH_TWOBARSXML = "/DuplicateNamesTest/TwoBars.xml";
    private String RES_PATH_TWOVALUESXML = "/DuplicateNamesTest/TwoValues.xml";
    private String RES_PATH_SEQUENCEABAXML = "/DuplicateNamesTest/SequenceABA.xml";
    private String RES_PATH_CHOICEABXML = "/DuplicateNamesTest/ChoiceAB.xml";

    private String NS_DUPLICATES = "http://www.ibm.com/example/duplicateNames";
    private String NS_DUPLICATESIMPORT = "http://www.ibm.com/example/duplicateNamesImport";
    
    // Test methods

    /**
     * Test name clash between global and anonymous types (OASIS issue SDO-82)
     */
    @Test
    public void testAnonymousTypeClash() throws IOException
    {
        Type personType = getScope().getTypeHelper().getType(NS_DUPLICATES, "person");
        Assert.assertEquals("person", personType.getName());
        Assert.assertNotNull(personType.getProperty("firstName"));
        Assert.assertNotNull(personType.getProperty("lastName"));
        
        Type personType2 = getScope().getTypeHelper().getOpenContentProperty(NS_DUPLICATES, "person").getType();
        Assert.assertNotNull(personType2.getProperty("name"));

        Type personType3 = getScope().getTypeHelper().getType(NS_DUPLICATES, "personRoot").getProperty("person").getType();
        Assert.assertNotNull(personType3.getProperty("identifier"));
    }
    
    /**
     * Test name clash between elements within a sequence model group (OASIS NEW issue)
     */
    @Test
    public void testSequenceWithDuplicates() throws IOException
    {
        Type sequenceABAType = getScope().getTypeHelper().getType(NS_DUPLICATES, "SequenceABA");
        Assert.assertTrue(sequenceABAType.isSequenced());
        
        Assert.assertNotNull(sequenceABAType.getProperty("A"));
        Assert.assertNotNull(sequenceABAType.getProperty("B"));
        Assert.assertTrue(sequenceABAType.getProperty("A").isMany());
        Assert.assertFalse(sequenceABAType.getProperty("B").isMany());

        DataObject dataObject = getScope().getDataFactory().create(sequenceABAType);
        dataObject.getSequence().add("A", "temp1");
        dataObject.getSequence().add("B", "temp2");
        dataObject.getSequence().add("A", "temp3");

        Assert.assertEquals("temp1", dataObject.get("A[1]"));
        Assert.assertEquals("temp2", dataObject.get("B"));
        Assert.assertEquals("temp3", dataObject.get("A[2]"));
        
        dataObject.setString("A[1]", "First");
        dataObject.setString("B", "Hello");
        dataObject.setString("A[2]", "Second");
        
        String actual = getScope().getXMLHelper().save(dataObject, NS_DUPLICATES, "sequenceABA");
        //System.out.println(actual);
        Assert.assertTrue(XMLCompareUtil.compareXML(getClass().getResourceAsStream(RES_PATH_SEQUENCEABAXML), actual));
        
        DataObject dataObject2 = getScope().getXMLHelper().load(actual).getRootObject();
        Assert.assertEquals("First", dataObject2.get("A[1]"));
        Assert.assertEquals("Hello", dataObject2.get("B"));
        Assert.assertEquals("Second", dataObject2.get("A[2]"));
    }
    
    /**
     * Test name clash between elements within a choice model group (OASIS NEW issue)
     */
    @Test
    public void testChoiceWithDuplicates() throws IOException
    {
        Type choiceABType = getScope().getTypeHelper().getType(NS_DUPLICATES, "ChoiceAB");
        Assert.assertTrue(choiceABType.isSequenced());
        
        Assert.assertNotNull(choiceABType.getProperty("A"));
        Assert.assertNotNull(choiceABType.getProperty("B"));
        Assert.assertFalse(choiceABType.getProperty("A").isMany());
        Assert.assertFalse(choiceABType.getProperty("B").isMany());

        DataObject dataObject = getScope().getDataFactory().create(choiceABType);
        dataObject.set("B", "value2");
        dataObject.set("A", "value1");
    
        String actual = getScope().getXMLHelper().save(dataObject, NS_DUPLICATES, "choiceAB");
        //System.out.println(actual);
        Assert.assertTrue(XMLCompareUtil.compareXML(getClass().getResourceAsStream(RES_PATH_CHOICEABXML), actual));
        
        DataObject dataObject2 = getScope().getXMLHelper().load(actual).getRootObject();
        Assert.assertEquals("value1", dataObject2.get("A"));
        Assert.assertEquals("value2", dataObject2.get("B"));
    }
    
    /**
     * Test name clash between an XML attribute and element with the same name
     * (OASIS issue SDO-71)
     */
    @Test
    public void testAttributeElementClash() throws IOException 
    {
        Type twoFoosType = getScope().getTypeHelper().getType(NS_DUPLICATES, "TwoFoos");
        Assert.assertTrue(hasProperty(twoFoosType, "", "foo", "String", true));
        Assert.assertTrue(hasProperty(twoFoosType, "", "@foo", "Int", false));
        Assert.assertFalse(twoFoosType.getProperty("@foo").getName().equals("foo"));
        Assert.assertTrue(twoFoosType.getProperty("@foo").getAliasNames().contains("@foo"));

        DataObject dataObject = getScope().getDataFactory().create(twoFoosType);
        dataObject.setString("foo", "Hello");
        dataObject.setInt("@foo", 101);
        
        String actual = getScope().getXMLHelper().save(dataObject, NS_DUPLICATES, "twoFoos");
        //System.out.println(actual);
        Assert.assertTrue(XMLCompareUtil.compareXML(getClass().getResourceAsStream(RES_PATH_TWOFOOSXML), actual));
        
        DataObject dataObject2 = getScope().getXMLHelper().load(actual).getRootObject();
        Assert.assertEquals("Hello", dataObject2.get("foo"));
        Assert.assertEquals(101, dataObject2.get("@foo"));
    }
    
    /**
     * Test name clash between special "value" property and an attribute named "value"
     * (OASIS issue SDO-83)
     */
    @Test
    public void testValuePropertyClash() throws IOException
    {
        Type twoValuesType = getScope().getTypeHelper().getType(NS_DUPLICATES, "TwoValues");
        Assert.assertTrue(hasProperty(twoValuesType, "", "value", "String", true));
        Assert.assertTrue(hasProperty(twoValuesType, "", "@value", "Int", false));
        Assert.assertFalse(twoValuesType.getProperty("@value").getName().equals("value"));
        Assert.assertTrue(twoValuesType.getProperty("@value").getAliasNames().contains("@value"));

        DataObject dataObject = getScope().getDataFactory().create(twoValuesType);
        dataObject.setString("value", "Hello");
        dataObject.setInt("@value", 101);
        
        String actual = getScope().getXMLHelper().save(dataObject, NS_DUPLICATES, "twoValues");
        //System.out.println(actual);
        Assert.assertTrue(XMLCompareUtil.compareXML(getClass().getResourceAsStream(RES_PATH_TWOVALUESXML), actual));
        
        DataObject dataObject2 = getScope().getXMLHelper().load(actual).getRootObject();
        Assert.assertEquals("Hello", dataObject2.get("value"));
        Assert.assertEquals(101, dataObject2.get("@value"));
    }
    
    /**
     * Test name clash between properties with the same local name, but different URIs
     * (OASIS issue SDO-67)
     */
    @Test
    public void testPropertyURIClash() throws IOException 
    {
        Type twoBarsType = getScope().getTypeHelper().getType(NS_DUPLICATES, "TwoBars");
        Assert.assertTrue(hasProperty(twoBarsType, "", "bar", "String", true));
        Assert.assertTrue(hasProperty(twoBarsType, NS_DUPLICATESIMPORT, "bar", "Int", true));

        DataObject dataObject = getScope().getDataFactory().create(twoBarsType);
        dataObject.setString("bar", "Hello");
        dataObject.setInt("bar[namespace-uri()=\"" + NS_DUPLICATESIMPORT + "\"]", 101);
        
        String actual = getScope().getXMLHelper().save(dataObject, NS_DUPLICATES, "twoBars");
        //System.out.println(actual);
        Assert.assertTrue(XMLCompareUtil.compareXML(getClass().getResourceAsStream(RES_PATH_TWOBARSXML), actual));
        
        DataObject dataObject2 = getScope().getXMLHelper().load(actual).getRootObject();
        Assert.assertEquals("Hello", dataObject2.get("bar"));
        Assert.assertEquals(101, dataObject2.get("bar[namespace-uri()='" + NS_DUPLICATESIMPORT + "']"));
    }
    
    @Before
    public void setUp() throws Exception {
        scope = SDOUtil.createHelperContext();

        URL url = getClass().getResource(RES_PATH_DUPLICATESXSD);
        InputStream is = url.openStream();
        getScope().getXSDHelper().define(is, url.toString());
        is.close();
    }
    
    protected boolean hasProperty(Type containingType, String uri, String name, String type, boolean isElement)
    {
        List<Property> properties = containingType.getProperties();
        for (Property property : properties)
        {
            if (uri.equals(getScope().getXSDHelper().getNamespaceURI(property)) &&
                (name.equals(property.getName()) || property.getAliasNames().contains(name)) &&
                type.equals(property.getType().getName()) &&
                isElement == getScope().getXSDHelper().isElement(property)) return true;                        
        }
        return false;
    }
    
    protected HelperContext scope = null;
    public HelperContext getScope() {
        return scope;
    }
}
